(function() {
  var RecordType, Type, VectorType, bigint, sixty_four_bits;
  var __hasProp = Object.prototype.hasOwnProperty, __extends = function(child, parent) { for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; } function ctor() { this.constructor = child; } ctor.prototype = parent.prototype; child.prototype = new ctor; child.__super__ = parent.prototype; return child; };

  bigint = require("bigint");

  sixty_four_bits = bigint("FFFFFFFFFFFFFFFF", 16);

  exports.Type = Type = (function() {

    function Type(name, normalize) {
      this.name = name;
      this.normalize = normalize;
    }

    Type.prototype.toString = function() {
      return "<Type " + this.name + ">";
    };

    return Type;

  })();

  exports.VectorType = VectorType = (function() {

    __extends(VectorType, Type);

    function VectorType(name, elementType) {
      var _this = this;
      this.elementType = elementType;
      VectorType.__super__.constructor.call(this, name, function(val) {
        var el, _i, _len, _results;
        _results = [];
        for (_i = 0, _len = val.length; _i < _len; _i++) {
          el = val[_i];
          _results.push(_this.elementType.normalize(el));
        }
        return _results;
      });
    }

    return VectorType;

  })();

  exports.RecordType = RecordType = (function() {

    function RecordType(name, fields) {
      var Record;
      if (name == null) throw new Error("RecordType name required");
      if (fields == null) throw new Error("RecordType fields required");
      Record = (function() {

        function Record(vals) {
          var field, type, _ref;
          if (vals != null) {
            _ref = Record.fields;
            for (field in _ref) {
              type = _ref[field];
              if (field in vals) this[field] = type.normalize(vals[field]);
            }
          }
        }

        return Record;

      })();
      Record.name = name;
      Record.fields = fields;
      Record.normalize = function(val) {
        if (val instanceof Record) {
          return val;
        } else {
          return new Record(val);
        }
      };
      return Record;
    }

    return RecordType;

  })();

  exports.byte = new Type("byte", function(val) {
    return val & 0xFF;
  });

  exports.boolean = new Type("boolean", Boolean);

  exports.int = new Type("int", function(val) {
    return Number(val) & 0xFFFFFFFF;
  });

  exports.long = new Type("long", function(val) {
    return bigint(val).and(sixty_four_bits);
  });

  exports.float = new Type("float", Number);

  exports.double = new Type("double", Number);

  exports.ustring = new Type("ustring", function(val) {
    if (val != null) {
      return String(val);
    } else {
      return null;
    }
  });

  exports.buffer = new Type("buffer", function(val) {
    if (val instanceof Buffer) {
      return val;
    } else if (val != null) {
      return new Buffer(String(val));
    } else {
      return null;
    }
  });

  exports.vector = function(type) {
    var name, _ref;
    name = (_ref = type.name) != null ? _ref : "?";
    return new VectorType("vector<" + name + ">", type);
  };

}).call(this);
